﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SunFP
{
    /// <summary>
    /// Implementation of simple test cases for 16.16 fixed point operations.
    /// </summary>
    class Testcases
    {

        public static void RunTests()
        {
            RunAddTests();
            Console.Out.WriteLine();
            RunSubTests();
            Console.Out.WriteLine();
            RunMulTests();
            Console.Out.WriteLine();
            RunDivTests();
            Console.Out.WriteLine();
            RunDiv2Tests();
            Console.Out.WriteLine();

            RunFloorTests();
            Console.Out.WriteLine();
            RunCeilTests();
            Console.Out.WriteLine();
            RunRoundTests();
            Console.Out.WriteLine();
            RunAbsTests();

            Console.Out.WriteLine();
            RunToStringTests();

            Console.Out.WriteLine();
            RunComparisonTests();

            Console.Out.WriteLine();
            RunPowerTests();
            Console.Out.WriteLine();
            RunFactorialTests();

            Console.Out.WriteLine();
            RunSinTests();
            Console.Out.WriteLine();
            RunCosTests();
            Console.Out.WriteLine();
            RunTanTests();
            Console.Out.WriteLine();
            RunSinhTests();
            Console.Out.WriteLine();
            RunCoshTests();
            Console.Out.WriteLine();
            RunTanhTests();

            Console.Out.WriteLine();
            RunExpTests();
            Console.Out.WriteLine();
            RunSqrtTests();

            Console.Out.WriteLine();
            RunArcTanhTests();
            Console.Out.WriteLine();
            RunLogNTests();
        }

        private static void RunFloorTests()
        {
            Console.Out.Write("Floor():");
            AssertEqual(SunFp.Floor(0), 0);
            AssertEqual(SunFp.Floor(1), 1);
            AssertEqual(SunFp.Floor(-3), -3);
            AssertEqual(SunFp.Floor(1.25f), 1);
            AssertEqual(SunFp.Floor(10.99f), 10);
            AssertEqual(SunFp.Floor(-2.7f), -3);
            AssertEqual(SunFp.Floor(-32767.99f), -32768);
            AssertEqual(SunFp.Floor(2), 2);
            AssertEqual(SunFp.Floor(2.4f), 2);
            AssertEqual(SunFp.Floor(2.9f), 2);
            AssertEqual(SunFp.Floor(-2.7f), -3);
            AssertEqual(SunFp.Floor(-2), -2);
        }

        private static void RunCeilTests()
        {
            Console.Out.Write("Ceil():");
            AssertEqual(SunFp.Ceil(0), 0);
            AssertEqual(SunFp.Ceil(2), 2);
            AssertEqual(SunFp.Ceil(2.4f), 3);
            AssertEqual(SunFp.Ceil(2.9f), 3);
            AssertEqual(SunFp.Ceil(-2.7f), -2);
            AssertEqual(SunFp.Ceil(-2f), -2);
        }

        private static void RunRoundTests()
        {
            Console.Out.Write("Round():");
            AssertEqual(SunFp.Round(0), 0);
            AssertEqual(SunFp.Round(1.1f), 1);
            AssertEqual(SunFp.Round(1.4f), 1);
            AssertEqual(SunFp.Round(1.6f), 2);
            AssertEqual(SunFp.Round(2.5f), 3);
            AssertEqual(SunFp.Round(-2.5f), -2);
            AssertEqual(SunFp.RoundMidpoint(2.5f, SunFp.ERoundMidpoint.SMALLER), 2);
            AssertEqual(SunFp.RoundMidpoint(2.5f, SunFp.ERoundMidpoint.LARGER), 3);
            AssertEqual(SunFp.Round(2.6f), 3);
            AssertEqual(SunFp.Round(-4.6f), -5);
            AssertEqual(SunFp.Round(-4.45f), -4);
            AssertEqual(SunFp.RoundMidpoint(-4.5f, SunFp.ERoundMidpoint.SMALLER), -5);
            AssertEqual(SunFp.RoundMidpoint(-4.5f, SunFp.ERoundMidpoint.LARGER), -4);
        }

        private static void RunAbsTests()
        {
            Console.Out.Write("Abs():");
            AssertEqual(SunFp.Abs(5), 5);
            AssertEqual(SunFp.Abs(0), 0);
            AssertEqual(SunFp.Abs(0.1f), 0.1f, 0.001f);
            AssertEqual(SunFp.Abs(-0.5f), 0.5f);
            AssertEqual(SunFp.Abs(-11), 11);
            AssertEqual(SunFp.Abs(-12f), 12f);
        }

        private static void RunMulTests()
        {
            Console.Out.Write("Mul():");
            AssertEqual(SunFp.Mul(3.5f, 4f), 14);
            AssertEqual(SunFp.Mul(-3.5f, 4f), -14);
            AssertEqual(SunFp.Mul(3.5f, -4f), -14);
            AssertEqual(SunFp.Mul(-3.5f, -4f), 14);

            AssertEqual(SunFp.Mul(2, 10), 20);
            AssertEqual(SunFp.Mul(45, 45), 2025);
            AssertEqual(SunFp.Mul(78, 3.75f), 292.5f);
            AssertEqual(SunFp.Mul(0.5f, 2), 1);
            AssertEqual(SunFp.Mul(3.5f, 2.5f), 8.75f);
            AssertEqual(SunFp.Mul(0.4f, 1.4f), 0.56f, 0.001f);
            AssertEqual(SunFp.Mul(0.3f, -0.3f), -0.09f, 0.001f);
            AssertEqual(SunFp.Mul(0.78f, 0.78f), 0.6083f, 0.001f);

            AssertEqual(new SunFp(0.4f) * new SunFp(1.4f), 0.56f, 0.001f);
        }

        private static void RunDivTests()
        {
            Console.Out.Write("Div():");
            AssertEqual(SunFp.Div(5, 1), 5);
            AssertEqual(SunFp.Div(-3.5f, 1), -3.5f);
            AssertEqual(SunFp.Div(5, 2), 2.5f);
            AssertEqual(SunFp.Div(8.75f, 2.5f), 3.5f);
            AssertEqual(SunFp.Div(6.25f, 2.5f), 2.5f);
            AssertEqual(SunFp.Div(8.75f, -2.5f), -3.5f);
            AssertEqual(SunFp.Div(-8.75f, 2.5f), -3.5f);
            AssertEqual(SunFp.Div(-8.75f, -2.5f), 3.5f);
            AssertEqual(SunFp.Div(1, 3), 0.33f, 0.005f);
            AssertEqual(SunFp.Div(7.5f, 3f), 2.5f);
 
        }

        private static void RunDiv2Tests()
        {
            Console.Out.Write("Div2():");
            AssertEqual(SunFp.Div2(5, 1), 5);
            AssertEqual(SunFp.Div2(-3.5f, 1), -3.5f);
            AssertEqual(SunFp.Div2(5, 2), 2.5f);
            AssertEqual(SunFp.Div2(8.75f, 2.5f), 3.5f, 0.01f);
            AssertEqual(SunFp.Div2(8.75f, -2.5f), -3.5f, 0.01f);
            AssertEqual(SunFp.Div2(-8.75f, 2.5f), -3.5f, 0.01f);
            AssertEqual(SunFp.Div2(-8.75f, -2.5f), 3.5f, 0.01f);
            AssertEqual(SunFp.Div2(1, 3), 0.33f, 0.005f);

            AssertEqual(SunFp.Div(7.5f, 3f), 2.5f);
        }

        private static void RunAddTests()
        {
            Console.Out.Write("Add():");
            AssertEqual(SunFp.Add(5, 1), 6);
            AssertEqual(SunFp.Add(0, 4), 4);
            AssertEqual(SunFp.Add(-1, 3), 2);
            AssertEqual(SunFp.Add(5, -2), 3);
            AssertEqual(SunFp.Add(7, 0.25f), 7.25f);
            AssertEqual(SunFp.Add(0.5f, 1.25f), 1.75f);
            AssertEqual(SunFp.Add(-64.3f, 4.8f), -59.5f, 0.001f);
            AssertEqual(SunFp.Add(32767, 1), -32768);
            AssertEqual(SunFp.Add(-1, 1), 0);
            AssertEqual(SunFp.Add(33, -0.75f), 32.25f);
            AssertEqual(SunFp.Add(57, 0.75f), 57.75f);
            AssertEqual(SunFp.Add(57, 1.11f), 58.11f, 0.001f);
            AssertEqual(new SunFp(7) + new SunFp(0.25f), 7.25f);
        }

        private static void RunSubTests()
        {
            Console.Out.Write("Sub():");
            AssertEqual(SunFp.Sub(5, 1), 4);
            AssertEqual(SunFp.Sub(0, 4), -4);
            AssertEqual(SunFp.Sub(-1, 3), -4);
            AssertEqual(SunFp.Sub(5, -2), 7);
            AssertEqual(SunFp.Sub(7, 0.25f), 6.75f);
            AssertEqual(SunFp.Sub(0.5f, 1.25f), -0.75f);
            AssertEqual(SunFp.Sub(-64.3f, 4.8f), -69.1f, 0.001f);
            AssertEqual(SunFp.Sub(32767, 1), 32766);
            AssertEqual(SunFp.Sub(-1, 1), -2);
            AssertEqual(SunFp.Sub(33, -0.75f), 33.75f);
            AssertEqual(SunFp.Sub(57, 0.75f), 56.25f);
            AssertEqual(SunFp.Sub(57, 1.11f), 55.89f, 0.001f);
            AssertEqual(SunFp.Sub(-1, -1), 0);
            AssertEqual(SunFp.Sub(-32768, 1), 32767);
            AssertEqual(new SunFp(57) - new SunFp(0.75f), 56.25f);
        }

        private static void RunToStringTests()
        {
            Console.Out.Write("ToString():");
            AssertEqual(SunFp.ToString(5), "5");
            AssertEqual(SunFp.ToString(0), "0");
            AssertEqual(SunFp.ToString(1.5f), "1.5");
            AssertEqual(SunFp.ToString(32.75f), "32.75");
            AssertEqual(SunFp.ToString(32767), "32767");
            AssertEqual(SunFp.ToString(2.1f), "2.099", 3);
            AssertEqual(SunFp.ToString(-2), "-2");
            AssertEqual(SunFp.ToString(-1), "-1");
            AssertEqual(SunFp.ToString(1), "1");
            AssertEqual(SunFp.ToString(-32767), "-32767");
            AssertEqual(SunFp.ToString(-0.75f), "-0.75");
            AssertEqual(SunFp.ToString(-0.92f), "-0.919", 3);
            AssertEqual(SunFp.ToString(0.44f), "0.439", 3);
            AssertEqual(SunFp.ToString(0.83f), "0.829", 3);
            AssertEqual(SunFp.ToString(-1.25f), "-1.25");
            AssertEqual(SunFp.ToString(-1.75f), "-1.75");
            AssertEqual(SunFp.ToString(-2.25f), "-2.25");
            AssertEqual(SunFp.ToString(-2.5f), "-2.5");
            AssertEqual(SunFp.ToString(-6.23f), "-6.229", 3);
        }

        private static void RunComparisonTests()
        {
            Console.Out.Write("Comparison():");
            AssertEqual(new SunFp(2) > new SunFp(1.9f), true);
            AssertEqual(new SunFp(2) >= new SunFp(2f), true);
            AssertEqual(new SunFp(2) < new SunFp(2f), false);
            AssertEqual(new SunFp(-2) < new SunFp(-2f), false);
            AssertEqual(new SunFp(-2) <= new SunFp(-2f), true);
            AssertEqual(new SunFp(-2.04f) < new SunFp(-2f), true);
        }

        private static void RunPowerTests()
        {
            Console.Out.Write("Power():");
            AssertEqual(SunFp.Power(2, 3), 8);
            AssertEqual(SunFp.Power(10, 4), 10000);
            AssertEqual(SunFp.Power(-1.5f, 2), 2.25f);
            AssertEqual(SunFp.Power(-1.5f, 3), -3.375f);

            AssertEqual(SunFp.Power(2.5f, 4.5f), 61.763f, 0.03f);
            AssertEqual(SunFp.Power(-1.5f, 3), -3.375f, 0.01f);
            AssertEqual(SunFp.Power(10, 4), 10000f, 0.01f);
            AssertEqual(SunFp.Power(3.45f, 0.6f), 2.102f, 0.01f);
            AssertEqual(SunFp.Power(56.4f, 1.2f), 126.338f, 1.3f);
        }

        private static void RunFactorialTests()
        {
            Console.Out.Write("Factorial():");
            AssertEqual(SunFp.Factorial(4), 24);
            AssertEqual(SunFp.Factorial(6), 720);
            AssertEqual(SunFp.Factorial(0), 1);
            AssertEqual(SunFp.Factorial(1), 1);
            AssertEqual(SunFp.Factorial(2), 2);
        }

        private static void RunSinTests()
        {
            Console.Out.Write("Sin():");
            AssertEqual(SunFp.Sin(0), 0);
            AssertEqual(SunFp.Sin((float)(Math.PI / 6)), 0.5f, 0.01f); /* 30° */
            AssertEqual(SunFp.Sin((float)(Math.PI / 4)), (float)(1 / Math.Sqrt(2)), 0.01f); /* 45° */
            AssertEqual(SunFp.Sin((float)(Math.PI / 3)), (float)(Math.Sqrt(3) / 2), 0.01f); /* 60° */
            AssertEqual(SunFp.Sin((float)(Math.PI / 2)), 1f, 0.01f); /* 90° */
            AssertEqual(SunFp.Sin((float)(Math.PI)), 0f, 0.01f); /* 180° */

            AssertEqual(SunFp.Sin((float)(-Math.PI / 6)), -0.5f, 0.01f); /* -30° */
            AssertEqual(SunFp.Sin((float)(-Math.PI / 4)), (float)(-1 / Math.Sqrt(2)), 0.01f); /* -45° */
            AssertEqual(SunFp.Sin((float)(-Math.PI / 3)), (float)(-Math.Sqrt(3) / 2), 0.01f); /* -60° */
            AssertEqual(SunFp.Sin((float)(-Math.PI / 2)), -1f, 0.01f); /* -90° */
            AssertEqual(SunFp.Sin((float)(-Math.PI)), 0f, 0.01f); /* -180° */
        }

        private static void RunCosTests()
        {
            Console.Out.Write("Cos():");
            AssertEqual(SunFp.Cos(0), 1f, 0.01f);
            AssertEqual(SunFp.Cos((float)(Math.PI / 6)), (float)(Math.Sqrt(3) / 2), 0.01f); /* 30° */
            AssertEqual(SunFp.Cos((float)(Math.PI / 4)), (float)(1 / Math.Sqrt(2)), 0.01f); /* 45° */
            AssertEqual(SunFp.Cos((float)(Math.PI / 3)), 0.5f, 0.01f); /* 60° */
            AssertEqual(SunFp.Cos((float)(Math.PI / 2)), 0f, 0.01f); /* 90° */
            AssertEqual(SunFp.Cos((float)(Math.PI)), -1f, 0.018f); /* 180° */

            AssertEqual(SunFp.Cos((float)(-Math.PI / 6)), (float)(Math.Sqrt(3) / 2), 0.01f); /* -30° */
            AssertEqual(SunFp.Cos((float)(-Math.PI / 4)), (float)(1 / Math.Sqrt(2)), 0.01f); /* -45° */
            AssertEqual(SunFp.Cos((float)(-Math.PI / 3)), 0.5f, 0.01f); /* -60° */
            AssertEqual(SunFp.Cos((float)(-Math.PI / 2)), 0f, 0.01f); /* -90° */
            AssertEqual(SunFp.Cos((float)(-Math.PI)), -1f, 0.018f); /* -180° */
        }

        private static void RunTanTests()
        {
            Console.Out.Write("Tan():");
            AssertEqual(SunFp.Tan(0), 0f, 0.01f);
            AssertEqual(SunFp.Tan((float)(Math.PI / 6)), (float)(1 / Math.Sqrt(3)), 0.01f); /* 30° */
            AssertEqual(SunFp.Tan((float)(Math.PI / 4)), 1f, 0.01f); /* 45° */
            AssertEqual(SunFp.Tan((float)(Math.PI / 3)), (float)(Math.Sqrt(3)), 0.01f); /* 60° */
            AssertEqual(SunFp.Tan((float)(Math.PI * 5 / 12)), (float)(2 + Math.Sqrt(3)), 0.01f); /* 75° */

            AssertEqual(SunFp.Tan((float)(-Math.PI / 6)), -(float)(1 / Math.Sqrt(3)), 0.01f); /* -30° */
            AssertEqual(SunFp.Tan((float)(-Math.PI / 4)), -1f, 0.01f); /* -45° */
            AssertEqual(SunFp.Tan((float)(-Math.PI / 3)), -(float)(Math.Sqrt(3)), 0.01f); /* -60° */
            AssertEqual(SunFp.Tan((float)(-Math.PI * 5 / 12)), -(float)(2 + Math.Sqrt(3)), 0.01f); /* -75° */
        }

        private static void RunSinhTests()
        {
            Console.Out.Write("Sinh():");
            AssertEqual(SunFp.Sinh(0), 0f, 0.01f);
            AssertEqual(SunFp.Sinh(1), 1.17f, 0.01f);
            AssertEqual(SunFp.Sinh((float)(Math.PI / 2)), 2.30f, 0.01f);
            AssertEqual(SunFp.Sinh((float)(Math.PI)), 11.54f, 0.1f);
            AssertEqual(SunFp.Sinh(-1), -1.17f, 0.01f);
            AssertEqual(SunFp.Sinh(-(float)(Math.PI / 2)), -2.30f, 0.01f);
            AssertEqual(SunFp.Sinh(-(float)(Math.PI)), -11.54f, 0.1f);
        }

        private static void RunCoshTests()
        {
            Console.Out.Write("Cosh():");
            AssertEqual(SunFp.Cosh(0), 1f, 0.01f);
            AssertEqual(SunFp.Cosh((float)Math.PI / 4), 1.32f, 0.01f);
            AssertEqual(SunFp.Cosh((float)Math.PI / 2), 2.509f, 0.01f);
            AssertEqual(SunFp.Cosh((float)Math.PI), 11.59f, 0.07f);
            AssertEqual(SunFp.Cosh(0.1f), 1.005f, 0.01f);
            AssertEqual(SunFp.Cosh(1.0f), 1.543f, 0.01f);
            AssertEqual(SunFp.Cosh(1.4f), 2.15f, 0.01f);

            AssertEqual(SunFp.Cosh((float)-Math.PI / 4), 1.32f, 0.01f);
            AssertEqual(SunFp.Cosh((float)-Math.PI / 2), 2.509f, 0.01f);
            AssertEqual(SunFp.Cosh((float)-Math.PI), 11.59f, 0.07f);
        }

        private static void RunTanhTests()
        {
            Console.Out.Write("Tanh():");
            AssertEqual(SunFp.Tanh(0), 0f, 0.01f);
            AssertEqual(SunFp.Tanh((float)Math.PI / 4), 0.656f, 0.01f);
            AssertEqual(SunFp.Tanh((float)Math.PI / 2), 0.9171f, 0.01f);
            AssertEqual(SunFp.Tanh((float)Math.PI), 0.9962f, 0.07f);
        }

        private static void RunExpTests()
        {
            Console.Out.Write("Exp():");
            AssertEqual(SunFp.Exp(0), 1f, 0.01f);
            AssertEqual(SunFp.Exp(1), 2.71f, 0.01f);
            AssertEqual(SunFp.Exp(-1), 0.367f, 0.01f);
            AssertEqual(SunFp.Exp(2), 7.39f, 0.01f);
            AssertEqual(SunFp.Exp(4), 54.6f, 0.01f);
            AssertEqual(SunFp.Exp(5.6f), 270.426f, 0.3f);
            AssertEqual(SunFp.Exp(-2.4f), 0.09f, 0.001f);
        }

        private static void RunArcTanhTests()
        {
            Console.Out.Write("ArcTanh():");

            AssertEqual(SunFp.ArcTanh(0), 0f);
            AssertEqual(SunFp.ArcTanh(0.5f), 0.549f, 0.01f);
            AssertEqual(SunFp.ArcTanh(0.7f), 0.867f, 0.01f);
            AssertEqual(SunFp.ArcTanh(-0.6f), -0.693f, 0.01f);
            AssertEqual(SunFp.ArcTanh(0.90f), 1.47f, 0.08f);
        }

        private static void RunLogNTests()
        {
            Console.Out.Write("LogN():");

            AssertEqual(SunFp.LogN_ArcTanH(2.5f), 0.916f, 0.01f);
            AssertEqual(SunFp.LogN_ArcTanH(0.6f), -0.51f, 0.01f);
            AssertEqual(SunFp.LogN_ArcTanH(0.1f), -2.30f, 0.04f);
            AssertEqual(SunFp.LogN_ArcTanH(11f), 2.397f, 0.04f);

            AssertEqual(SunFp.LogN_Taylor(2.5f), 0.916f, 0.01f);
            AssertEqual(SunFp.LogN_Taylor(0.6f), -0.51f, 0.01f);
            AssertEqual(SunFp.LogN_Taylor(0.2f), -1.60f, 0.07f);
            AssertEqual(SunFp.LogN_Taylor(3.45f), 1.238f, 0.04f);
            AssertEqual(SunFp.LogN_Taylor(11f), 2.397f, 0.04f);
            AssertEqual(SunFp.LogN_Taylor(25f), 3.218f, 0.04f);
            AssertEqual(SunFp.LogN_Taylor(5641), 8.6378f, 0.04f);
            AssertEqual(SunFp.LogN_Taylor(25741.5f), 10.1558f, 0.04f);
        }

        private static void RunSqrtTests()
        {
            Console.Out.Write("Sqrt():");

            AssertEqual(SunFp.Sqrt(2), 1.414f, 0.001f);
            AssertEqual(SunFp.Sqrt(9), 3);
            AssertEqual(SunFp.Sqrt(100), 10);
            AssertEqual(SunFp.Sqrt(6.25f), 2.5f);
        }

        /*********************************************************************
         * Assertion functions
         *********************************************************************/

        private static void AssertEqual(bool actualBool, bool expectedBool)
        {
            if (actualBool.Equals(expectedBool))
            {
                Console.Out.Write('x');
            }
            else
            {
                Console.Out.Write("F\n");
                Console.Out.WriteLine("Error: Value " + actualBool + " - Expected: " + expectedBool);
            }
        }

        private static void AssertEqual(string fpStr, string expectedStr)
        {
            if (fpStr.Equals(expectedStr))
            {
                Console.Out.Write('x');
            }
            else
            {
                Console.Out.Write("F\n");
                Console.Out.WriteLine("Error: Value " + fpStr + " - Expected: " + expectedStr);
            }
        }

        private static void AssertEqual(string fpStr, string expectedStr, int numFracDigits)
        {
            string truncStr = fpStr.Substring(0, fpStr.IndexOf('.') + numFracDigits + 1);

            if (truncStr.Equals(expectedStr))
            {
                Console.Out.Write('x');
            }
            else
            {
                Console.Out.Write("F\n");
                Console.Out.WriteLine("Error: Value " + truncStr + " - Expected: " + expectedStr);
            }
        }

        private static void AssertEqual(SunFp fp, float expectedVal, float epsilon)
        {
            if (Math.Abs(fp.ToFloat() - expectedVal) <= epsilon)
            {
                Console.Out.Write('x');
            }
            else
            {
                Console.Out.Write("F\n");
                Console.Out.WriteLine("Error: Value " + fp.ToFloat() + " - Expected: " + expectedVal);
            }
        }

        private static void AssertEqual(SunFp fp, float expectedVal)
        {
            if (fp.ToFloat() == expectedVal)
            {
                Console.Out.Write('x');
            }
            else
            {
                Console.Out.Write("F\n");
                Console.Out.WriteLine("Error: Value " + fp.ToFloat() + " - Expected: " + expectedVal);
            }
        }

        private static void AssertEqual(SunFp fp, int expectedVal)
        {
            if (fp.ToInt() == expectedVal)
            {
                Console.Out.Write('x');
            }
            else
            {
                Console.Out.Write("F\n");
                Console.Out.WriteLine("Error: Value " + fp.ToInt() + " - Exepcted: " + expectedVal);
            }
        }
    }
}
